@@ -1,23 +1,26 @@
-/*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2013-2018 Antonio Diaz Diaz.
+/* GNU ddrescue - Data recovery tool
+   Copyright (C) 2013-2023 Antonio Diaz Diaz.
 
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 2 of the License, or
-    (at your option) any later version.
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 2 of the License, or
+   (at your option) any later version.
 
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #define _FILE_OFFSET_BITS 64
 
+#include <algorithm>
 #include <cstdio>
+#include <cstdlib>
+#include <cstring>
 #include <string>
 #include <vector>
 #include <sys/stat.h>
@@ -28,15 +31,15 @@
 
 namespace {
 
-const char * format_time_dhms( const long t )
+const char * format_time_dhms( const long long t )
   {
-  static char buf[32];
+  static char buf[64];			// keep gcc quiet
   const int s = t % 60;
   const int m = ( t / 60 ) % 60;
   const int h = ( t / 3600 ) % 24;
-  const long d = t / 86400;
+  const long long d = t / 86400;
 
-  if( d ) snprintf( buf, sizeof buf, "%ldd:%02dh:%02dm:%02ds", d, h, m, s );
+  if( d ) snprintf( buf, sizeof buf, "%lldd:%02dh:%02dm:%02ds", d, h, m, s );
   else if( h ) snprintf( buf, sizeof buf, "%dh:%02dm:%02ds", h, m, s );
   else if( m ) snprintf( buf, sizeof buf, "%dm:%02ds", m, s );
   else snprintf( buf, sizeof buf, "%ds", s );
@@ -51,22 +54,24 @@ Rate_logger rate_logger;
 Read_logger read_logger;
 
 
-bool Logger::set_filename( const char * const name )
+void Logger::set_filename( const char * const name )
   {
-  if( name && name[0] )
-    {
-    struct stat st;
-    if( stat( name, &st ) == 0 && !S_ISREG( st.st_mode ) ) return false;
-    filename_ = name;
-    }
-  return true;
+  if( !name || !name[0] ) return;			// ignore name
+  if( std::strcmp( name, "-" ) == 0 )
+    { show_error( "I won't write log data to standard output." );
+      std::exit( 1 ); }
+  struct stat st;
+  if( stat( name, &st ) == 0 && !S_ISREG( st.st_mode ) )
+    { show_file_error( name, "Logfile exists and is not a regular file." );
+      std::exit( 1 ); }
+  filename_ = name;
   }
 
 
 bool Logger::close_file()
   {
   if( f && !error && !write_final_timestamp( f ) ) error = true;
-  if( f && std::fclose( f ) != 0 ) error = true;
+  if( f && std::fclose( f ) != 0 ) error = true;	// close even on error
   f = 0;
   return !error;
   }
@@ -97,7 +102,8 @@ bool Event_logger::echo_msg( const char * const msg )
   }
 
 
-bool Event_logger::print_msg( const long time, const char * const percent_rescued,
+bool Event_logger::print_msg( const long long time,
+                              const char * const percent_rescued,
                               const char * const msg )
   {
   if( f && !error &&
@@ -108,7 +114,8 @@ bool Event_logger::print_msg( const long time, const char * const percent_rescue
   }
 
 
-bool Event_logger::print_eor( const long time, const char * const percent_rescued,
+bool Event_logger::print_eor( const long long time,
+                              const char * const percent_rescued,
                               const long long current_pos,
                               const char * const current_status_name )
   {
@@ -135,7 +142,7 @@ bool Rate_logger::open_file()
   }
 
 
-bool Rate_logger::print_line( const long time, const long long ipos,
+bool Rate_logger::print_line( const long long time, const long long ipos,
                               const long long a_rate, const long long c_rate,
                               const unsigned long bad_areas,
                               const long long bad_size )
@@ -143,7 +150,7 @@ bool Rate_logger::print_line( const long time, const long long ipos,
   if( f && !error && time > last_time )
     {
     last_time = time;
-    if( std::fprintf( f, "%2ld  0x%08llX  %8lld  %8lld  %7lu  %8lld\n",
+    if( std::fprintf( f, "%2lld  0x%08llX  %8lld  %8lld  %7lu  %8lld\n",
                       time, ipos, c_rate, a_rate, bad_areas, bad_size ) < 0 )
       error = true;
     }
@@ -177,7 +184,7 @@ bool Read_logger::print_line( const long long ipos, const long long size,
   }
 
 
-bool Read_logger::print_msg( const long time, const char * const msg )
+bool Read_logger::print_msg( const long long time, const char * const msg )
   {
   if( f && !error &&
       std::fprintf( f, "%s# %s  %s\n", prev_is_msg ? "" : "\n",
@@ -188,7 +195,7 @@ bool Read_logger::print_msg( const long time, const char * const msg )
   }
 
 
-bool Read_logger::print_time( const long time )
+bool Read_logger::print_time( const long long time )
   {
   if( f && !error && time > 0 &&
       std::fprintf( f, "# %s\n", format_time_dhms( time ) ) < 0 )
