@@ -1,20 +1,20 @@
-/*  Rational - Rational number class with overflow detection
-    Copyright (C) 2005-2018 Antonio Diaz Diaz.
+/* Rational - Rational number class with overflow detection
+   Copyright (C) 2005-2023 Antonio Diaz Diaz.
 
-    This library is free software. Redistribution and use in source and
-    binary forms, with or without modification, are permitted provided
-    that the following conditions are met:
+   This library is free software. Redistribution and use in source and
+   binary forms, with or without modification, are permitted provided
+   that the following conditions are met:
 
-    1. Redistributions of source code must retain the above copyright
-    notice, this list of conditions and the following disclaimer.
+   1. Redistributions of source code must retain the above copyright
+   notice, this list of conditions, and the following disclaimer.
 
-    2. Redistributions in binary form must reproduce the above copyright
-    notice, this list of conditions and the following disclaimer in the
-    documentation and/or other materials provided with the distribution.
+   2. Redistributions in binary form must reproduce the above copyright
+   notice, this list of conditions, and the following disclaimer in the
+   documentation and/or other materials provided with the distribution.
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+   This library is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
 
 #include <algorithm>
@@ -26,10 +26,10 @@
 #include "rational.h"
 
 #ifndef LLONG_MAX
-#define LLONG_MAX  0x7FFFFFFFFFFFFFFFLL
+#define LLONG_MAX 0x7FFFFFFFFFFFFFFFLL
 #endif
 #ifndef LLONG_MIN
-#define LLONG_MIN  (-LLONG_MAX - 1LL)
+#define LLONG_MIN (-LLONG_MAX - 1LL)
 #endif
 #ifndef ULLONG_MAX
 #define ULLONG_MAX 0xFFFFFFFFFFFFFFFFULL
@@ -51,7 +51,7 @@ long long gcd( long long n, long long m )	// Greatest Common Divisor
   }
 
 
-const std::string overflow_string( const int n )
+std::string overflow_string( const int n )
   { if( n > 0 ) return "+INF"; if( n < 0 ) return "-INF"; return "NAN"; }
 
 int overflow_value( const long long n, const bool negate = false )
@@ -118,10 +118,10 @@ Rational & Rational::operator+=( const Rational & r )
   if( den <= 0 ) return *this;			// no op on error
   if( r.den <= 0 ) { num = r.num; den = 0; return *this; }	// set error
 
-  const long long new_den = (long long)den * r.den;
-  const long long new_num = ( (long long)num * r.den ) +
-                            ( (long long)r.num * den );
-  normalize( new_num, new_den );
+  long long new_den = den; new_den *= r.den;
+  long long new_num1 = num; new_num1 *= r.den;
+  long long new_num2 = r.num; new_num2 *= den;
+  normalize( new_num1 + new_num2, new_den );
   return *this;
   }
 
@@ -131,8 +131,8 @@ Rational & Rational::operator*=( const Rational & r )
   if( den <= 0 ) return *this;			// no op on error
   if( r.den <= 0 ) { num = r.num; den = 0; return *this; }	// set error
 
-  const long long new_num = (long long)num * r.num;
-  const long long new_den = (long long)den * r.den;
+  long long new_num = num; new_num *= r.num;
+  long long new_den = den; new_den *= r.den;
   normalize( new_num, new_den );
   return *this;
   }
@@ -149,11 +149,11 @@ int Rational::round() const
   }
 
 
-// Recognized formats: 123 123/456 123.456 .123 12% 12/3% 12.3% .12%
-// Values may be preceded by an optional '+' or '-' sign.
-// Returns the number of chars read from 's', or 0 if input is invalid.
-// In case of invalid input, the Rational is not changed.
-//
+/* Recognized formats: 123 123/456 123.456 .123 12% 12/3% 12.3% .12%
+   Values may be preceded by an optional '+' or '-' sign.
+   Return the number of chars read from 's', or 0 if input is invalid.
+   In case of invalid input, the Rational is not changed.
+*/
 int Rational::parse( const char * const s )
   {
   if( !s || !s[0] ) return 0;
@@ -209,13 +209,15 @@ int Rational::parse( const char * const s )
   }
 
 
-// Returns a string representing the value 'num/den' in decimal point
-// format with 'prec' decimals.
-// 'iwidth' is the minimum width of the integer part, prefixed with
-// spaces if needed.
-// If 'prec' is negative, only the needed decimals are produced.
-//
-const std::string Rational::to_decimal( const unsigned iwidth, int prec ) const
+/* Return a string representing the value 'num/den' in decimal point format
+   with 'prec' decimals.
+   'iwidth' is the minimum width of the integer part, prefixed with spaces
+   if needed.
+   If 'prec' is negative, produce only the decimals needed.
+   If 'rounding', round up the last digit if the next one would be >= 5.
+*/
+std::string Rational::to_decimal( const unsigned iwidth, int prec,
+                                  const bool rounding ) const
   {
   if( den <= 0 ) return overflow_string( num );
 
@@ -224,7 +226,7 @@ const std::string Rational::to_decimal( const unsigned iwidth, int prec ) const
   const bool truncate = ( prec < 0 );
   if( prec < 0 ) prec = -prec;
 
-  do { s += '0' + ( ipart % 10 ); ipart /= 10; } while( ipart > 0 );
+  do { s += ( ipart % 10 ) + '0'; ipart /= 10; } while( ipart > 0 );
   if( num < 0 ) s += '-';
   if( iwidth > s.size() ) s.append( iwidth - s.size(), ' ' );
   std::reverse( s.begin(), s.end() );
@@ -233,26 +235,40 @@ const std::string Rational::to_decimal( const unsigned iwidth, int prec ) const
     {
     s += '.';
     while( prec > 0 && ( rest > 0 || !truncate ) )
-      { rest *= 10; s += '0' + ( rest / den ); rest %= den; --prec; }
+      { rest *= 10; s += ( rest / den ) + '0'; rest %= den; --prec; }
     }
+  if( rounding && rest * 2 >= den )		// round last decimal up
+    for( int j = s.size() - 1; j >= 0; --j )
+      {
+      if( s[j] == '.' ) continue;
+      if( s[j] >= '0' && s[j] < '9' ) { ++s[j]; break; }
+      if( s[j] == '9' ) s[j] = '0';
+      if( j > 0 && s[j-1] == '.' ) continue;
+      if( j > 0 && s[j-1] == ' ' ) { s[j-1] = '1'; break; }
+      if( j > 1 && s[j-2] == ' ' && s[j-1] == '-' )
+        { s[j-2] = '-'; s[j-1] = '1'; break; }
+      // no prev digit, prepend '1' to the first digit
+      if( j == 0 || s[j-1] < '0' || s[j-1] > '9' )
+        { s.insert( s.begin() + j, '1' ); break; }
+      }
   return s;
   }
 
 
-// Returns a string representing the value 'num/den' in fractional form.
-// 'width' is the minimum width to be produced, prefixed with spaces if
-// needed.
-//
-const std::string Rational::to_fraction( const unsigned width ) const
+/* Return a string representing the value 'num/den' in fractional form.
+   'width' is the minimum width to be produced, prefixed with spaces if
+   needed.
+*/
+std::string Rational::to_fraction( const unsigned width ) const
   {
   if( den <= 0 ) return overflow_string( num );
 
   std::string s;
   int n = std::abs( num ), d = den;
 
-  do { s += '0' + ( d % 10 ); d /= 10; } while( d > 0 );
+  do { s += ( d % 10 ) + '0'; d /= 10; } while( d > 0 );
   s += '/';
-  do { s += '0' + ( n % 10 ); n /= 10; } while( n > 0 );
+  do { s += ( n % 10 ) + '0'; n /= 10; } while( n > 0 );
   if( num < 0 ) s += '-';
   if( width > s.size() ) s.append( width - s.size(), ' ' );
   std::reverse( s.begin(), s.end() );
