@@ -1,18 +1,18 @@
-/*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004-2018 Antonio Diaz Diaz.
+/* GNU ddrescue - Data recovery tool
+   Copyright (C) 2004-2023 Antonio Diaz Diaz.
 
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 2 of the License, or
-    (at your option) any later version.
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 2 of the License, or
+   (at your option) any later version.
 
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #define _FILE_OFFSET_BITS 64
@@ -94,8 +94,14 @@ bool Rescuebook::extend_outfile_size()
     if( size < 0 ) return false;
     if( min_size > size )
       {
-      const uint8_t zero = 0;
-      if( writeblockp( odes_, &zero, 1, min_size - 1 ) != 1 ) return false;
+      int ret;
+      do ret = ftruncate( odes_, min_size );
+        while( ret != 0 && errno == EINTR );
+      if( ret != 0 || lseek( odes_, 0, SEEK_END ) != min_size )
+        {
+        const uint8_t zero = 0;		// if ftruncate fails, write a zero
+        if( writeblockp( odes_, &zero, 1, min_size - 1 ) != 1 ) return false;
+        }
       fsync( odes_ );
       }
     }
@@ -103,9 +109,9 @@ bool Rescuebook::extend_outfile_size()
   }
 
 
-// Return values: 2 bad infile, 1 I/O error, 0 OK.
-// If OK && copied_size + error_size < b.size(), it means EOF has been reached.
-//
+/* Return values: 1 fatal error, 0 OK (I/O errors are ignored).
+   If OK && copied_size + error_size < b.size(), it means EOF has been reached.
+*/
 int Rescuebook::copy_block( const Block & b, int & copied_size, int & error_size )
   {
   if( b.size() <= 0 ) internal_error( "bad size copying a Block." );
@@ -127,8 +133,17 @@ int Rescuebook::copy_block( const Block & b, int & copied_size, int & error_size
       }
     else copied_size = readblockp( ides_, iobuf(), b.size(), b.pos() );
     error_size = errno ? b.size() - copied_size : 0;
-    if( errno == EINVAL )
-      { final_msg( "Unaligned read error. Is sector size correct?" ); return 1; }
+    if( copied_size <= 0 ) switch( errno )
+      {
+      case EACCES: case EBADF: case EBUSY: case EISDIR: case ENOBUFS:
+      case ENODEV: case ENOENT: case ENOMEM: case ENOSYS: case ENXIO:
+      case EPERM: case ESPIPE:
+        final_msg( iname_, "Fatal error reading the input file", errno );
+        return 1;
+      case EINVAL:
+        final_msg( iname_, "Unaligned read error. Is sector size correct?" );
+        return 1;
+      }
     }
   else { copied_size = 0; error_size = b.size(); }
 
@@ -141,9 +156,13 @@ int Rescuebook::copy_block( const Block & b, int & copied_size, int & error_size
       const long long end = pos + copied_size;
       if( end > sparse_size ) sparse_size = end;
       }
-    else if( writeblockp( odes_, iobuf(), copied_size, pos ) != copied_size ||
-             ( synchronous_ && fsync( odes_ ) != 0 && errno != EINVAL ) )
-      { final_msg( "Write error", errno ); return 1; }
+    else
+      if( !compare_before_write ||
+          readblockp( odes_, iobuf2(), copied_size, pos ) != copied_size ||
+          std::memcmp( iobuf(), iobuf2(), copied_size ) != 0 )
+        if( writeblockp( odes_, iobuf(), copied_size, pos ) != copied_size ||
+              ( synchronous_ && fsync( odes_ ) != 0 && errno != EINVAL ) )
+          { final_msg( oname_, "Write error", errno ); return 1; }
     }
   else iobuf_ipos = -1;
 
@@ -157,14 +176,16 @@ int Rescuebook::copy_block( const Block & b, int & copied_size, int & error_size
       {
       if( voe_ipos >= 0 )
         {
-        const int size = readblockp( ides_, iobuf_aux(), hardbs(), voe_ipos );
+        const int size = readblockp( ides_, iobuf2(), hardbs(), voe_ipos );
         if( size != hardbs() )
-          { final_msg( "Input file no longer returns data", errno ); e_code |= 8; }
-        else if( std::memcmp( voe_buf, iobuf_aux(), hardbs() ) != 0 )
-          { final_msg( "Input file returns inconsistent data" ); e_code |= 8; }
+          { final_msg( iname_, "Input file no longer returns data", errno );
+            e_code |= 8; }
+        else if( std::memcmp( voe_buf, iobuf2(), hardbs() ) != 0 )
+          { final_msg( iname_, "Input file returns inconsistent data." );
+            e_code |= 8; }
         }
       else
-        { final_msg( "Read error found before the first good read" );
+        { final_msg( iname_, "Read error found before the first good read." );
           e_code |= 8; }
       }
     }
@@ -206,12 +227,11 @@ int Rescuebook::copy_and_update( const Block & b, int & copied_size,
   {
   if( first_post )
     {
-    if( first_read ) first_read = false;
-    else if( pause_on_pass > 0 )
+    if( !first_read && pause_on_pass > 0 )
       {
       show_status( -1, "Paused", true );
       sleep( pause_on_pass );
-      const long t2 = std::time( 0 );
+      const long long t2 = std::time( 0 );
       if( t1 < t2 ) t1 = t2;			// clock may have jumped back
       ts = std::min( ts + pause_on_pass, t2 );	// avoid spurious timeout
       }
@@ -231,7 +251,7 @@ int Rescuebook::copy_and_update( const Block & b, int & copied_size,
       {
       if( complete_only ) truncate_domain( b.pos() + copied_size + error_size );
       else if( !truncate_vector( b.pos() + copied_size + error_size ) )
-        { final_msg( "EOF found below the size calculated from mapfile" );
+        { final_msg( iname_, "EOF found below the size calculated from mapfile." );
           retval = 1; }
       initialize_sizes();
       }
@@ -247,16 +267,16 @@ int Rescuebook::copy_and_update( const Block & b, int & copied_size,
       change_chunk_status( Block( b.pos() + copied_size, error_size ), st2 );
       struct stat istat;
       if( stat( iname_, &istat ) != 0 )
-        { final_msg( "Input file disappeared", errno ); retval = 1; }
+        { final_msg( iname_, "Input file disappeared", errno ); retval = 1; }
       }
     }
   return retval;
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Read the non-tried part of the domain, skipping over the damaged areas.
-//
+/* Return values: 1 error, 0 OK, -1 interrupted, -2 mapfile error.
+   Read the non-tried part of the domain, skipping over the damaged areas.
+*/
 int Rescuebook::copy_non_tried()
   {
   char msgbuf[80] = "Copying non-tried blocks... Pass ";
@@ -286,10 +306,10 @@ int Rescuebook::copy_non_tried()
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Read forwards the non-tried part of the domain, skipping over the
-// damaged areas.
-//
+/* Return values: 1 error, 0 OK, -1 interrupted, -2 mapfile error.
+   Read forwards the non-tried part of the domain, skipping over the
+   damaged areas.
+*/
 int Rescuebook::fcopy_non_tried( const char * const msg, const int pass,
                                  const bool resume )
   {
@@ -298,6 +318,7 @@ int Rescuebook::fcopy_non_tried( const char * const msg, const int pass,
   long long sskip_size = 0;		// size to skip on slow if skipbs > 0
   const bool after_finished = (pass == 3 || pass == 4 );
   bool block_found = false;
+  bool block_processed = false;
 
   if( resume && domain().includes( current_pos() ) )
     {
@@ -312,6 +333,7 @@ int Rescuebook::fcopy_non_tried( const char * const msg, const int pass,
     if( find_chunk( b, Sblock::non_tried, domain(), softbs(), after_finished ) )
       block_found = true;
     if( b.size() <= 0 ) break;
+    block_processed = true;
     if( pos != b.pos() )		// reset size on block change
       { eskip_size = skipbs; current_slow = false; }
     pos = b.end();
@@ -354,14 +376,15 @@ int Rescuebook::fcopy_non_tried( const char * const msg, const int pass,
     if( !update_mapfile( odes_ ) ) return -2;
     }
   if( !block_found ) return 0;
+  if( block_processed ) show_status( -1, msg, true );	// update at end of pass
   return -3;
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Read backwards the non-tried part of the domain, skipping over the
-// damaged areas.
-//
+/* Return values: 1 error, 0 OK, -1 interrupted, -2 mapfile error.
+   Read backwards the non-tried part of the domain, skipping over the
+   damaged areas.
+*/
 int Rescuebook::rcopy_non_tried( const char * const msg, const int pass,
                                  const bool resume )
   {
@@ -370,6 +393,7 @@ int Rescuebook::rcopy_non_tried( const char * const msg, const int pass,
   long long sskip_size = 0;		// size to skip on slow if skipbs > 0
   const bool before_finished = (pass == 3 || pass == 4 );
   bool block_found = false;
+  bool block_processed = false;
 
   if( resume && domain().includes( current_pos() - 1 ) )
     {
@@ -384,6 +408,7 @@ int Rescuebook::rcopy_non_tried( const char * const msg, const int pass,
     if( rfind_chunk( b, Sblock::non_tried, domain(), softbs(), before_finished ) )
       block_found = true;
     if( b.size() <= 0 ) break;
+    block_processed = true;
     if( end != b.end() )		// reset size on block change
       { eskip_size = skipbs; current_slow = false; }
     end = b.pos();
@@ -425,14 +450,15 @@ int Rescuebook::rcopy_non_tried( const char * const msg, const int pass,
     if( !update_mapfile( odes_ ) ) return -2;
     }
   if( !block_found ) return 0;
+  if( block_processed ) show_status( -1, msg, true );	// update at end of pass
   return -3;
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Trim both edges of each damaged area sequentially. If any edge is
-// adjacent to a bad sector, leave it for the scraping phase.
-//
+/* Return values: 1 error, 0 OK, -1 interrupted, -2 mapfile error.
+   Trim both edges of each damaged area sequentially. If any edge is
+   adjacent to a bad sector, leave it for the scraping phase.
+*/
 int Rescuebook::trim_errors()
   {
   const char * const msg = reverse ? "Trimming failed blocks... (backwards)" :
@@ -442,7 +468,7 @@ int Rescuebook::trim_errors()
   for( long i = 0; i < sblocks(); )
     {
     const long idx = reverse ? sblocks() - 1 - i : i;
-    const Sblock sb = sblock( idx );
+    const Sblock sb( sblock( idx ) );
     if( !domain().includes( sb ) )
       { if( ( !reverse && domain() < sb ) || ( reverse && domain() > sb ) )
           break;
@@ -495,13 +521,14 @@ int Rescuebook::trim_errors()
         change_chunk_status( sblock( index ), Sblock::non_scraped );
       }
     }
+  show_status( -1, msg, true );			// update at end of pass
   return 0;
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Scrape the damaged areas sequentially.
-//
+/* Return values: 1 error, 0 OK, -1 interrupted, -2 mapfile error.
+   Scrape the damaged areas sequentially.
+*/
 int Rescuebook::scrape_errors()
   {
   const char * const msg = reverse ? "Scraping failed blocks... (backwards)" :
@@ -510,7 +537,7 @@ int Rescuebook::scrape_errors()
 
   for( long i = 0; i < sblocks(); )
     {
-    const Sblock sb = sblock( reverse ? sblocks() - 1 - i : i );
+    const Sblock sb( sblock( reverse ? sblocks() - 1 - i : i ) );
     if( !domain().includes( sb ) )
       { if( ( !reverse && domain() < sb ) || ( reverse && domain() > sb ) )
           break;
@@ -532,13 +559,14 @@ int Rescuebook::scrape_errors()
       if( !update_mapfile( odes_ ) ) return -2;
       }
     }
+  show_status( -1, msg, true );			// update at end of pass
   return 0;
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Try to read the damaged areas, one sector at a time.
-//
+/* Return values: 1 error, 0 OK, -1 interrupted, -2 mapfile error.
+   Try to read the damaged areas, one sector at a time.
+*/
 int Rescuebook::copy_errors()
   {
   char msgbuf[80] = "Retrying bad sectors... Retry ";
@@ -566,9 +594,9 @@ int Rescuebook::copy_errors()
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Try to read forwards the damaged areas, one sector at a time.
-//
+/* Return values: 1 error, 0 OK, -1 interrupted, -2 mapfile error.
+   Try to read forwards the damaged areas, one sector at a time.
+*/
 int Rescuebook::fcopy_errors( const char * const msg, const int pass,
                               const bool resume )
   {
@@ -598,13 +626,14 @@ int Rescuebook::fcopy_errors( const char * const msg, const int pass,
     if( !update_mapfile( odes_ ) ) return -2;
     }
   if( !block_found ) return 0;
+  show_status( -1, msg, true );			// update at end of pass
   return -3;
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Try to read backwards the damaged areas, one sector at a time.
-//
+/* Return values: 1 error, 0 OK, -1 interrupted, -2 mapfile error.
+   Try to read backwards the damaged areas, one sector at a time.
+*/
 int Rescuebook::rcopy_errors( const char * const msg, const int pass,
                               const bool resume )
   {
@@ -634,11 +663,12 @@ int Rescuebook::rcopy_errors( const char * const msg, const int pass,
     if( !update_mapfile( odes_ ) ) return -2;
     }
   if( !block_found ) return 0;
+  show_status( -1, msg, true );			// update at end of pass
   return -3;
   }
 
 
-// Returns true if slow read.
+// Return true if slow read.
 //
 bool Rescuebook::update_rates( const bool force )
   {
@@ -655,12 +685,12 @@ bool Rescuebook::update_rates( const bool force )
       }
     }
 
-  long t2 = std::time( 0 );
+  long long t2 = std::time( 0 );
   if( max_read_rate > 0 && finished_size - last_size > max_read_rate && t2 == t1 )
     { sleep( 1 ); t2 = std::time( 0 ); }
   if( t2 < t1 )					// clock jumped back
     {
-    const long delta = std::min( t0 - 1, t1 - t2 );
+    const long long delta = std::min( t0 - 1, t1 - t2 );
     t0 -= delta;
     ts -= delta;
     t1 = t2;
@@ -671,7 +701,7 @@ bool Rescuebook::update_rates( const bool force )
     {
     if( tp > 0 )
       {
-      const long delta = std::min( t0 - 1, (long)tp.round() );
+      const long long delta = std::min( t0 - 1, (long long)tp.round() );
       t0 -= delta;
       t1 -= delta;
       ts -= delta;
@@ -717,10 +747,13 @@ void Rescuebook::show_status( const long long ipos, const char * const msg,
     {
     if( verbosity >= 0 )
       {
-      std::printf( "\r%s%s%s%s%s%s", up, up, up, up, up, up );
+      if( first_post && !first_read && !rescue_finished() )
+        std::fputc( '\n', stdout );	// scroll forward after each pass
+      else std::printf( "\r%s%s%s%s%s%s", up, up, up, up, up, up );
       if( preview_lines > 0 )
         {
-        for( int i = -2; i < preview_lines; ++i ) std::fputs( up, stdout );
+        if( !first_post || first_read )
+          for( int i = -2; i < preview_lines; ++i ) std::fputs( up, stdout );
         std::fputs( "Data preview:\n", stdout );
         for( int i = 0; i < preview_lines; ++i )
           {
@@ -775,25 +808,26 @@ void Rescuebook::show_status( const long long ipos, const char * const msg,
         for( int i = len; i < oldlen; ++i ) std::fputc( ' ', stdout );
         oldlen = len;
         }
-      std::fflush( stdout );
+      safe_fflush( stdout );
       }
     rate_logger.print_line( t1 - t0, last_ipos, a_rate, c_rate, bad_areas,
                             bad_size );
     if( !force && !first_post ) read_logger.print_time( t1 - t0 );
     rates_updated = false;
     first_post = false;
+    first_read = false;
     }
   }
 
 
-Rescuebook::Rescuebook( const long long offset, const long long isize,
+Rescuebook::Rescuebook( const long long offset, const long long insize,
                         Domain & dom, const Domain * const test_dom,
                         const Mb_options & mb_opts, const Rb_options & rb_opts,
-                        const char * const iname, const char * const mapname,
-                        const int cluster, const int hardbs,
-                        const bool synchronous )
-  : Mapbook( offset, isize, dom, mb_opts, mapname, cluster, hardbs,
-             rb_opts.complete_only ),
+                        const char * const iname, const char * const oname,
+                        const char * const mapname, const int cluster,
+                        const int hardbs, const bool synchronous )
+  : Mapbook( offset, insize, dom, mb_opts, mapname, cluster, hardbs,
+             rb_opts.complete_only, true ),
     Rb_options( rb_opts ),
     error_rate( 0 ),
     error_sum( 0 ),
@@ -804,7 +838,7 @@ Rescuebook::Rescuebook( const long long offset, const long long isize,
     bad_size( 0 ),
     finished_size( 0 ),
     test_domain( test_dom ),
-    iname_( iname ),
+    iname_( iname ), oname_( oname ),
     read_errors( 0 ),
     slow_reads( 0 ),
     e_code( 0 ),
@@ -813,15 +847,15 @@ Rescuebook::Rescuebook( const long long offset, const long long isize,
     a_rate( 0 ), c_rate( 0 ), first_size( 0 ), last_size( 0 ),
     iobuf_ipos( -1 ), last_ipos( 0 ), t0( 0 ), t1( 0 ), ts( 0 ), tp( 0 ),
     oldlen( 0 ), rates_updated( false ), current_slow( false ),
-    prev_slow( false ), sliding_avg( 30 ), first_post( false ),
+    prev_slow( false ), sliding_avg( 30 ), first_post( true ),
     first_read( true )
   {
   if( preview_lines > softbs() / 16 ) preview_lines = softbs() / 16;
   if( skipbs < 0 )
-    skipbs = round_up( std::max( isize / 100000, (long long)min_skipbs ),
+    skipbs = round_up( std::max( insize / 100000, (long long)min_skipbs ),
                        min_skipbs );
-  const long long csize = isize / 100;
-  if( isize > 0 && skipbs > 0 && max_skipbs == max_max_skipbs &&
+  const long long csize = insize / 100;
+  if( insize > 0 && skipbs > 0 && max_skipbs == max_max_skipbs &&
       csize < max_skipbs )
     max_skipbs = std::max( skipbs, csize );
   skipbs = round_up( skipbs, hardbs );		// make multiple of hardbs
@@ -852,16 +886,11 @@ Rescuebook::Rescuebook( const long long offset, const long long isize,
   }
 
 
-// Return values: 1 I/O error, 0 OK.
+// Return values: 0 OK, != 0 error.
 //
 int Rescuebook::do_rescue( const int ides, const int odes )
   {
-  bool copy_pending = false, trim_pending = false, scrape_pending = false;
   ides_ = ides; odes_ = odes;
-
-  if( non_tried_size ) copy_pending = trim_pending = scrape_pending = true;
-  if( non_trimmed_size )              trim_pending = scrape_pending = true;
-  if( non_scraped_size )                             scrape_pending = true;
   set_signals();
   if( verbosity >= 0 )
     {
@@ -877,9 +906,10 @@ int Rescuebook::do_rescue( const int ides, const int odes )
           std::printf( " last block size: %9sB\n",
                        format_num( sblock( sblocks() - 1 ).size() ) );
         }
-      if( domain().pos() > 0 || domain().end() < mapfile_isize() )
-        std::printf( "(sizes limited to domain %lld B to %lld B of %lld B)\n",
-                     domain().pos(), domain().end(), mapfile_isize() );
+      if( domain().pos() > 0 || domain().end() < mapfile_insize() )
+        std::printf( "(sizes limited to domain from %s B to %s B of %s B)\n",
+                     format_num3( domain().pos() ), format_num3( domain().end() ),
+                     format_num3( mapfile_insize() ) );
       std::printf( "rescued: %sB, tried: %sB, bad-sector: %sB, bad areas: %lu\n\n",
                    format_num( finished_size ),
                    format_num( non_trimmed_size + non_scraped_size + bad_size ),
@@ -889,16 +919,16 @@ int Rescuebook::do_rescue( const int ides, const int odes )
     }
   int retval = 0;
   update_rates();				// first call
-  if( copy_pending && !errors_or_timeout() )
+  if( non_tried_size && !errors_or_timeout() )
     retval = copy_non_tried();
-  if( retval == 0 && trim_pending && !notrim && !errors_or_timeout() )
+  if( retval == 0 && non_trimmed_size && !notrim && !errors_or_timeout() )
     retval = trim_errors();
-  if( retval == 0 && scrape_pending && !noscrape && !errors_or_timeout() )
+  if( retval == 0 && non_scraped_size && !noscrape && !errors_or_timeout() )
     retval = scrape_errors();
-  if( retval == 0 && max_retries != 0 && !errors_or_timeout() )
+  if( retval == 0 && bad_size && max_retries != 0 && !errors_or_timeout() )
     retval = copy_errors();
   if( !rates_updated ) update_rates( true );	// force update of e_code
-  show_status( -1, retval ? 0 : "Finished", true );
+  show_status( -1, retval ? 0 : "\nFinished", true );
 
   const bool signaled = ( retval == -1 );
   if( signaled ) retval = 0;
@@ -923,7 +953,7 @@ int Rescuebook::do_rescue( const int ides, const int odes )
       if( e_code & 32 ) event_logger.echo_msg( "Too many slow reads" );
       }
     if( verbosity >= 0 )
-      { std::fputc( '\n', stdout ); std::fflush( stdout ); }
+      { std::fputc( '\n', stdout ); safe_fflush( stdout ); }
     }
   if( retval == -2 ) retval = 1;		// mapfile error
   else
@@ -931,26 +961,27 @@ int Rescuebook::do_rescue( const int ides, const int odes )
     if( retval == 0 && !signaled ) current_status( finished );
     if( !extend_outfile_size() )		// sparse or -x option
       {
-      show_error( "Error extending output file size." );
+      show_file_error( oname_, "Error extending output file size." );
       if( retval == 0 ) retval = 1;
       }
     compact_sblock_vector();
     if( !update_mapfile( odes_, true ) && retval == 0 ) retval = 1;
     }
-  if( final_msg().size() )
-    { if( final_errno() ) show_error( final_msg().c_str(), final_errno() );
-      else { std::fputs( final_msg().c_str(), stdout ); std::fputc( '\n', stdout ); } }
+  if( final_msg().size() ) show_error( final_msg().c_str(), final_errno() );
   if( close( odes_ ) != 0 )
-    { show_error( "Error closing outfile", errno );
+    { show_file_error( oname_, "Error closing outfile", errno );
       if( retval == 0 ) retval = 1; }
   event_logger.print_eor( t1 - t0, percent_rescued(), current_pos(),
                           status_name( current_status() ) );
   if( !event_logger.close_file() )
-    show_error( "warning: Error closing the events logging file." );
+    show_file_error( event_logger.filename(),
+                     "warning: Error closing the events logging file." );
   if( !rate_logger.close_file() )
-    show_error( "warning: Error closing the rates logging file." );
+    show_file_error( rate_logger.filename(),
+                     "warning: Error closing the rates logging file." );
   if( !read_logger.close_file() )
-    show_error( "warning: Error closing the reads logging file." );
+    show_file_error( read_logger.filename(),
+                     "warning: Error closing the reads logging file." );
   if( retval ) return retval;		// errors have priority over signals
   if( signaled ) return signaled_exit();
   return 0;
